# pylint: disable=C0103,C0111,C0302,W0212
import unittest
import pytest

import quasardb


def test_connect_throws_input_error__when_uri_is_invalid():
    with pytest.raises(quasardb.Error):
        quasardb.Cluster(uri='invalid_uri')


def test_connect_throws_connection_error_when_no_cluster_on_given_uri():
    with pytest.raises(quasardb.Error):
        quasardb.Cluster(uri='qdb://127.0.0.1:1')


def test_connect_throws_connection_error_when_no_cluster_public_key(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Cluster(
            uri=qdbd_settings.get("uri").get("secure"),
            user_name=qdbd_settings.get("security").get("user_name"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"))


def test_connect_throws_connection_error_when_no_user_name(qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Cluster(
            uri=qdbd_settings.get("uri").get("secure"),
            user_private_key=qdbd_settings.get("security").get("user_private_key"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


def test_connect_throws_connection_error_when_no_user_private_key(
        qdbd_settings):
    with pytest.raises(quasardb.Error):
        quasardb.Cluster(
            uri=qdbd_settings.get("uri").get("secure"),
            user_name=qdbd_settings.get("security").get("user_name"),
            cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


def test_connect_ok_to_secure_cluster(qdbd_settings):
    quasardb.Cluster(
        uri=qdbd_settings.get("uri").get("secure"),
        user_name=qdbd_settings.get("security").get("user_name"),
        user_private_key=qdbd_settings.get("security").get("user_private_key"),
        cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))



from __future__ import print_function

import argparse
import sys
import sysconfig

from . import get_include


def print_includes():
    dirs = [sysconfig.get_path('include'),
            sysconfig.get_path('platinclude'),
            get_include(),
            get_include(True)]

    # Make unique but preserve order
    unique_dirs = []
    for d in dirs:
        if d not in unique_dirs:
            unique_dirs.append(d)

    print(' '.join('-I' + d for d in unique_dirs))


def main():
    parser = argparse.ArgumentParser(prog='python -m pybind11')
    parser.add_argument('--includes', action='store_true',
                        help='Include flags for both pybind11 and Python headers.')
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.includes:
        print_includes()


if __name__ == '__main__':
    main()



# Copyright (c) 2009-2019, quasardb SAS
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function
from builtins import range as xrange, int

import os
from socket import gethostname
import sys
import inspect
import traceback
import random
import time
import datetime
import locale
import numpy as np

import quasardb

STOCK_COLUMN = "stock_id"
OPEN_COLUMN = "open"
CLOSE_COLUMN = "close"
HIGH_COLUMN = "high"
LOW_COLUMN = "low"
VOLUME_COLUMN = "volume"

def time_execution(str, f, *args):
    print("     - ", str, end='')

    start_time = time.time()
    res = f(*args)
    end_time = time.time()

    print(" [duration: {}s]".format(end_time - start_time))

    return res

def gen_ts_name():
    return "test.{}.{}.{}".format(gethostname(), os.getpid(), random.randint(0, 100000))

def create_ts(q, name):
    ts = q.ts(name)
    ts.create([quasardb.ColumnInfo(quasardb.ColumnType.Int64, STOCK_COLUMN),
               quasardb.ColumnInfo(quasardb.ColumnType.Double, OPEN_COLUMN),
               quasardb.ColumnInfo(quasardb.ColumnType.Double, CLOSE_COLUMN),
               quasardb.ColumnInfo(quasardb.ColumnType.Double, HIGH_COLUMN),
               quasardb.ColumnInfo(quasardb.ColumnType.Double, LOW_COLUMN),
               quasardb.ColumnInfo(quasardb.ColumnType.Int64, VOLUME_COLUMN)])

    return ts

def create_many_ts(q, names):
    return [create_ts(q, x) for x in names]

def generate_prices(price_count):
    return np.random.uniform(-100.0, 100.0, price_count)

def generate_points(points_count):
    start_time = np.datetime64('2017-01-01', 'ns')

    dates = np.array([(start_time + np.timedelta64(i, 'm')) for i in range(points_count)]).astype('datetime64[ns]')
    stock_ids = np.random.randint(1, 25, size=points_count)
    prices = np.array([generate_prices(60) for i in range(points_count)]).astype('double')
    volumes = np.random.randint(0, 10000, points_count)

    return (dates, stock_ids, prices, volumes)

def batch_ts_columns(ts_name, prealloc_size):
    return (quasardb.BatchColumnInfo(ts_name, STOCK_COLUMN, prealloc_size),
            quasardb.BatchColumnInfo(ts_name, OPEN_COLUMN, prealloc_size),
            quasardb.BatchColumnInfo(ts_name, CLOSE_COLUMN, prealloc_size),
            quasardb.BatchColumnInfo(ts_name, HIGH_COLUMN, prealloc_size),
            quasardb.BatchColumnInfo(ts_name, LOW_COLUMN, prealloc_size),
            quasardb.BatchColumnInfo(ts_name, VOLUME_COLUMN, prealloc_size))

def calculate_minute_bar(prices):
    # Takes all prices for a single minute, and calculate OHLC
    return (prices[0], prices[-1], np.amax(prices), np.amin(prices))

def bulk_insert(q, ts_names, dates, stock_ids, prices, volumes):
    # We generate a flattened list of columns for each timeseries; for example,
    # for 2 columns for 4 timeseries each, we have 8 columns.
    columns = [column for nested in (batch_ts_columns(ts_name, len(dates))
                                     for ts_name in ts_names)
               for column in nested]

    batch_inserter = q.ts_batch(columns)
    for i in range(len(stock_ids)):
        # We use the known layout of column (2 for each timeseries, alternating with
        # STOCK_COLUMN and PRICE_COLUMN) to set the values.
        for j in range(0, len(ts_names) * 6, 6):
            (o, c, h, l) = calculate_minute_bar(prices[i])

            batch_inserter.start_row(dates[i])
            batch_inserter.set_int64(j, stock_ids[i]) # set stock_id
            batch_inserter.set_double(j + 1, o) # open
            batch_inserter.set_double(j + 2, c) # close
            batch_inserter.set_double(j + 3, h) # high
            batch_inserter.set_double(j + 4, l) # low
            batch_inserter.set_int64(j + 5, volumes[i]) # low

    batch_inserter.push()

def make_it_so(q, points_count):
    ts_names = [gen_ts_name(), gen_ts_name()]

    ts = time_execution("Creating a time series with names {}".format(ts_names), create_many_ts, q, ts_names)
    (dates, stock_ids, prices, volumes) = time_execution("Generating {:,} points".format(points_count), generate_points, points_count)
    time_execution("Inserting {:,} points into timeseries with names {}".format(points_count, ts_names), bulk_insert, q, ts_names, dates, stock_ids, prices, volumes)

    return (ts_names, dates, np.unique(stock_ids))

def main(quasardb_uri, points_count):
    print("Connecting to: ", quasardb_uri)
    q = quasardb.Cluster(uri=quasardb_uri)

    print(" *** Inserting {:,} into {}".format(points_count, quasardb_uri))
    make_it_so(q, points_count)

if __name__ == "__main__":

    try:
        if len(sys.argv) != 3:
            print("usage: ", sys.argv[0], " quasardb_uri points_count")
            sys.exit(1)

        main(sys.argv[1], int(sys.argv[2]))

    except Exception as ex:  # pylint: disable=W0703
        print("An error ocurred:", str(ex))
        traceback.print_exc()



# pylint: disable=C0103,C0111,C0302,W0212
import random
import string
import pytest
import numpy as np

import quasardb


def connect(uri):
    return quasardb.Cluster(uri)


def config():
    return {"uri":
            {"insecure": "qdb://127.0.0.1:28360",
             "secure": "qdb://127.0.0.1:28361"}}


@pytest.fixture
def qdbd_settings(scope="module"):
    return {
        "uri": {
            "insecure": "qdb://127.0.0.1:28360",
            "secure": "qdb://127.0.0.1:28361"},
        "security": {
            "user_name": 'qdb-api-python',
            "user_private_key": 'SoHHpH26NtZvfq5pqm/8BXKbVIkf+yYiVZ5fQbq1nbcI=',
            "cluster_public_key": 'Pb+d1o3HuFtxEb5uTl9peU89ze9BZTK9f8KdKr4k7zGA='}}


@pytest.fixture
def qdbd_connection(qdbd_settings):
    return connect(qdbd_settings.get("uri").get("insecure"))


@pytest.fixture
def random_identifier():
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(16))


@pytest.fixture
def random_string():
    return ''.join(
        random.choice(
            string.ascii_uppercase +
            string.digits) for _ in range(16))


@pytest.fixture
def column_name(random_identifier):
    return random_identifier


@pytest.fixture
def tag_name(random_identifier):
    return random_identifier


@pytest.fixture
def tag_names():
    return sorted([''.join(random.choice(string.ascii_lowercase)
                           for _ in range(16)) for _ in range(10)])


@pytest.fixture
def entry_name(random_string):
    return random_string


@pytest.fixture
def random_blob(random_string):
    return random_string.encode('UTF-8')


@pytest.fixture
def random_integer():
    return random.randint(-1000000000, 1000000000)


@pytest.fixture
def blob_entry(qdbd_connection, entry_name):
    return qdbd_connection.blob(entry_name)


@pytest.fixture
def integer_entry(qdbd_connection, entry_name):
    return qdbd_connection.integer(entry_name)


@pytest.fixture
def table(qdbd_connection, entry_name):

    ts = qdbd_connection.ts(entry_name)
    double_col = quasardb.ColumnInfo(quasardb.ColumnType.Double, "the_double")
    blob_col = quasardb.ColumnInfo(quasardb.ColumnType.Blob, "the_blob")
    int64_col = quasardb.ColumnInfo(quasardb.ColumnType.Int64, "the_int64")
    ts_col = quasardb.ColumnInfo(quasardb.ColumnType.Timestamp, "the_ts")

    ts.create([double_col, blob_col, int64_col, ts_col])
    return ts


@pytest.fixture
def intervals():
    start_time = np.datetime64('2017-01-01', 'ns')
    return [(start_time, start_time + np.timedelta64(1, 's'))]


@pytest.fixture
def many_intervals():
    start_time = np.datetime64('2017-01-01', 'ns')
    return np.array([(start_time + np.timedelta64(i, 's'))
                     for i in range(10000)]).astype('datetime64[ns]')



# Copyright (c) 2009-2019, quasardb SAS
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from __future__ import print_function

import tempfile
import shutil
import os
import traceback
import sys
import subprocess
import time

from zipfile import ZipFile

import quasardb  # pylint: disable=C0413,E0401

dji_tickers = frozenset(["mmm", "axp", "aapl", "ba", "cat", "cvx", "csco", "ko", "dis", "xom",
                         "ge",  "gs",  "hd",   "ibm","intc","jnj", "jpm",  "mcd","mrk", "msft",
                         "nke", "pfe", "pg",   "trv","utx", "unh", "vz",   "v",  "wmt", "dd"])

nyse_tickers = frozenset(["mmm", "axp", "ba", "cat", "cvx", "ko", "dis", "xom",
                         "ge",  "gs",  "hd",   "ibm", "jnj", "jpm",  "mcd","mrk",
                         "nke", "pfe", "pg",   "trv","utx", "unh", "vz",   "v",  "wmt", "dd"])

nasdaq_tickers = frozenset(["aapl", "csco", "intc", "msft"])

def trim_suffix(str, suffix):
    idx = str.find(suffix)
    return str if idx == -1 else str[:idx]

def clean_name(company_name):
    cleaned = company_name.lower()
    cleaned = trim_suffix(cleaned, ' group')
    cleaned = trim_suffix(cleaned, ' corp')
    cleaned = trim_suffix(cleaned, ' inc')
    cleaned = trim_suffix(cleaned, ' co.')
    cleaned = trim_suffix(cleaned, ' llc')
    cleaned = trim_suffix(cleaned, ' limited')
    cleaned = trim_suffix(cleaned, '.com')
    cleaned = cleaned.strip()
    cleaned = cleaned.replace(',', '').replace('.', '').replace('&', '').replace("'", '').replace('-', '_').replace(')', '').replace('(', '').replace(' ', '_')
    cleaned = cleaned.replace('__', '_')

    return cleaned

def get_company_name(ticker_name):
    lookup_table = {"mmm": "3m",
                    "axp": "american express",
                    "aapl": "apple",
                    "ba": "boeing",
                    "cat": "caterpillar",
                    "cvx": "chevron",
                    "csco": "cisco",
                    "ko": "coca-cola",
                    "dis": "disney",
                    "dwdp": "dowdupont inc",
                    "xom": "exxon mobil",
                    "ge": "general electric",
                    "gs": "goldman sachs",
                    "hd": "home depot",
                    "ibm": "ibm",
                    "intc": "intel",
                    "jnj": "johnson and johnson",
                    "jpm": "jpmorgan chase",
                    "mcd": "mcdonald",
                    "mrk": "merck",
                    "msft": "microsoft",
                    "nke": "nike",
                    "pfe": "pfizer",
                    "pg": "procter and gamble",
                    "trv": "travelers companies inc",
                    "utx": "united technologies",
                    "unh": "unitedhealth",
                    "vz": "verizon",
                    "v": "visa",
                    "wmt": "walmart"}

    if ticker_name in lookup_table:
        return lookup_table[ticker_name]

    return ticker_name

def browse_zips(directory):
    for (dirpath, dirnames, filenames) in os.walk(directory):
        for filename in filenames:
            if filename.endswith('.zip'):
                yield os.path.join(dirpath, filename)

def insert_via_binary(quasardb_uri, file, key_name):
    subprocess.check_call(["qdb_insert_csv_cpp", quasardb_uri, file, key_name])

def remove_entry(q, key_name):
    try:
        ts = q.ts(key_name)
        ts.remove()
    except Exception as e:
        pass

def tag_entry(q, key_name, ticker_name):

    tickers = [ticker_name]

    if ticker_name in dji_tickers:
        tickers.append("dji")
        tickers.append("usd")

    if ticker_name in nyse_tickers:
        tickers.append("nyse")

    if ticker_name in nasdaq_tickers:
        tickers.append("nasdaq")

    print('Tagging ' + key_name + ' with ' + str(tickers))

    ts = q.ts(key_name)

    for t in tickers:
        ts.attach_tag(t)

def display_elapsed(start_time):
    print("...done in {} seconds".format(time.time() - start_time))

def main(quasardb_uri, directory):

    q = quasardb.Cluster(quasardb_uri)

    work_dir = tempfile.mkdtemp(prefix='qdb')

    for zip_file in browse_zips(directory):
        with ZipFile(zip_file, 'r') as zf:
            for f in zf.namelist():
                ticker_name = os.path.splitext(f)[0].lower()
                try:
                    start_time = time.time()

                    key_name = 'stocks.' + clean_name(get_company_name(ticker_name))

                    print('Importing ticker {} ({}) into {}...'.format(ticker_name, get_company_name(ticker_name), key_name))
                    zf.extract(f, work_dir)

                    full_path = os.path.join(work_dir, f)

                    remove_entry(q, ticker_name)

                    insert_via_binary(quasardb_uri, full_path, key_name)

                    tag_entry(q, key_name, ticker_name)

                    os.remove(full_path)

                    display_elapsed(start_time)

                except Exception as ex:
                    print('Could not insert {}: {}'.format(ticker_name, str(ex)))

    shutil.rmtree(work_dir)

if __name__ == "__main__":

    try:
        if len(sys.argv) != 3:
            print("usage: ", sys.argv[0], " quasardb_uri directory")
            sys.exit(1)

        main(sys.argv[1], sys.argv[2])

    except Exception as ex:  # pylint: disable=W0703
        print("An error ocurred:", str(ex))
        traceback.print_exc()


# pylint: disable=C0103,C0111,C0302,W0212

import quasardb
import pytest


def test_set_client_max_in_buf_1(qdbd_connection):
    with pytest.raises(quasardb.Error):
        qdbd_connection.options().set_client_max_in_buf_size(1)


def test_set_client_max_in_buf_1MB(qdbd_connection):
    qdbd_connection.options().set_client_max_in_buf_size(1024 * 1024)
    assert qdbd_connection.options().get_client_max_in_buf_size() == (1024 * 1024)


def test_get_client_max_in_buf(qdbd_connection):
    assert qdbd_connection.options().get_client_max_in_buf_size() > 0


def test_get_cluster_max_in_buf(qdbd_connection):
    assert qdbd_connection.options().get_cluster_max_in_buf_size() > 0



# # pylint: disable=C0103,C0111,C0302,W0212
import pytest
import quasardb
import numpy as np

import test_ts as tslib


def _insert_double_points(table, start_time, points=10):
    inserted_double_data = tslib._generate_double_ts(start_time, points)
    table.double_insert(tslib._double_col_name(table),
                        inserted_double_data[0],
                        inserted_double_data[1])
    return inserted_double_data


def _insert_blob_points(table, start_time, points=10):
    inserted_blob_data = tslib._generate_blob_ts(start_time, points)
    table.blob_insert(tslib._blob_col_name(table),
                      inserted_blob_data[0],
                      inserted_blob_data[1])
    return inserted_blob_data


def _insert_int64_points(table, start_time, points=10):
    inserted_int64_data = tslib._generate_int64_ts(start_time, points)
    table.int64_insert(tslib._int64_col_name(table),
                       inserted_int64_data[0],
                       inserted_int64_data[1])
    return inserted_int64_data


def _insert_timestamp_points(table, start_time, points=10):
    inserted_timestamp_data = tslib._generate_timestamp_ts(
        start_time, start_time, points)
    table.timestamp_insert(tslib._ts_col_name(table),
                           inserted_timestamp_data[0],
                           inserted_timestamp_data[1])
    return inserted_timestamp_data

##
# Query failure tests


def test_returns_invalid_argument_for_null_query(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.query()


def test_returns_invalid_argument_for_empty_query(qdbd_connection):
    q = qdbd_connection.query('')
    with pytest.raises(quasardb.Error):
        q.run()


def test_returns_invalid_argument_with_invalid_query(qdbd_connection):
    q = qdbd_connection.query('select * from')
    with pytest.raises(quasardb.Error):
        q.run()


def test_returns_alias_not_found_when_ts_doesnt_exist(qdbd_connection):
    q = qdbd_connection.query(
        'select * from ' + 'this_ts_doesnt_exist' + ' in range(2017, +10d)')
    with pytest.raises(quasardb.Error):
        q.run()


def test_returns_alias_not_found_when_untagged(qdbd_connection, tag_name):
    q = qdbd_connection.query(
        "select * from find(tag='" + tag_name + "') in range(2017, +10d)")
    with pytest.raises(quasardb.Error):
        q.run()


def test_returns_columns_not_found(qdbd_connection, table, column_name):
    q = qdbd_connection.query(
        "select " +
        column_name +
        " from " +
        table.get_name() +
        " in range(2017, +10d)")
    with pytest.raises(quasardb.Error):
        q.run()


##
# Double data tests

def sanity_check(ts_name, scanned_point_count, res, rows_count, columns_count):
    assert res.scanned_point_count == scanned_point_count
    assert len(res.tables) == 1
    assert len(res.tables[ts_name][0].data) == rows_count
    assert len(res.tables[ts_name]) == columns_count
    assert res.tables[ts_name][0].name == "$timestamp"


def test_returns_empty_result(qdbd_connection, table):
    res = qdbd_connection.query(
        "select * from " +
        table.get_name() +
        " in range(2016-01-01 , 2016-12-12)").run()
    assert res.scanned_point_count == 0
    assert len(res.tables) == 0


def test_returns_inserted_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from " + table.get_name() + \
        " in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    # Column count is 5, because, uninit, int64, blob, timestamp, double
    sanity_check(table.get_name(),
                 len(inserted_double_data[0]),
                 res,
                 len(inserted_double_data[0]),
                 5)

    assert res.tables[table.get_name()][0].name == "$timestamp"
    np.testing.assert_array_equal(
        res.tables[table.get_name()][0].data, inserted_double_data[0])
    assert res.tables[table.get_name(
    )][1].name == tslib._double_col_name(table)
    np.testing.assert_array_equal(
        res.tables[table.get_name()][1].data, inserted_double_data[1])


def test_returns_inserted_data_with_star_select_and_tag_lookup(
        qdbd_connection, table, tag_name, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    table.attach_tag(tag_name)
    query = "select * from find(tag = " + '"' + tag_name + '")' + \
        " in range(" + str(tslib._start_year(intervals)) + ", +100d)"

    res = qdbd_connection.query(query).run()
    sanity_check(table.get_name(),
                 len(inserted_double_data[0]),
                 res,
                 len(inserted_double_data[0]),
                 5)

    assert res.tables[table.get_name()][0].name == "$timestamp"
    np.testing.assert_array_equal(
        res.tables[table.get_name()][0].data, inserted_double_data[0])

    assert res.tables[table.get_name(
    )][1].name == tslib._double_col_name(table)
    np.testing.assert_array_equal(
        res.tables[table.get_name()][1].data, inserted_double_data[1])


def test_returns_inserted_data_with_column_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select " + tslib._double_col_name(table) + " from " + table.get_name(
    ) + " in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    sanity_check(table.get_name(),
                 len(inserted_double_data[0]),
                 res,
                 len(inserted_double_data[0]),
                 2)

    assert res.tables[table.get_name()][0].name == "$timestamp"
    np.testing.assert_array_equal(
        res.tables[table.get_name()][0].data, inserted_double_data[0])
    assert res.tables[table.get_name(
    )][1].name == tslib._double_col_name(table)
    np.testing.assert_array_equal(
        res.tables[table.get_name()][1].data, inserted_double_data[1])


def test_returns_inserted_data_twice_with_double_column_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select " + tslib._double_col_name(table) + "," + tslib._double_col_name(table) + \
        " from " + table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()
    sanity_check(table.get_name(), len(inserted_double_data[0]),
                 res, len(inserted_double_data[0]), 2)

    assert res.tables[table.get_name()][0].name == "$timestamp"
    np.testing.assert_array_equal(
        res.tables[table.get_name()][0].data, inserted_double_data[0])
    assert res.tables[table.get_name(
    )][1].name == tslib._double_col_name(table)
    np.testing.assert_array_equal(
        res.tables[table.get_name()][1].data, inserted_double_data[1])


def test_returns_sum_with_sum_select(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select sum(" + tslib._double_col_name(table) + ") from " + table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()
    sanity_check(table.get_name(), len(inserted_double_data[0]), res, 1, 2)

    res_table = res.tables[table.get_name()]
    assert res_table[0].name == "$timestamp"
    assert np.isnat(res_table[0].data[0])
    assert res_table[1].name == "sum(" + tslib._double_col_name(table) + ")"
    assert pytest.approx(res_table[1].data[0], 0.1) == np.sum(
        inserted_double_data[1])


def test_returns_sum_with_sum_divided_by_count_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select sum(" + tslib._double_col_name(table) + ")/count(" + \
        tslib._double_col_name(table) + ") from " + table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()
    sanity_check(table.get_name(), len(inserted_double_data[0]) * 2, res, 1, 2)

    res_table = res.tables[table.get_name()]
    assert res_table[0].name == "$timestamp"
    assert np.isnat(res_table[0].data[0])
    assert res_table[1].name == "(sum(" + tslib._double_col_name(
        table) + ")/count(" + tslib._double_col_name(table) + "))"
    assert pytest.approx(res_table[1].data[0], 0.1) == np.average(
        inserted_double_data[1])


def test_returns_max_minus_min_select(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select max(" + tslib._double_col_name(table) + ") - min(" + \
        tslib._double_col_name(table) + ") from " + table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    sanity_check(table.get_name(), len(inserted_double_data[0]) * 2, res, 1, 2)

    res_table = res.tables[table.get_name()]
    assert res_table[0].name == "$timestamp"
    assert np.isnat(res_table[0].data[0])
    assert res_table[1].name == "(max(" + tslib._double_col_name(
        table) + ")-min(" + tslib._double_col_name(table) + "))"
    assert pytest.approx(res_table[1].data[0], 0.1) == np.max(
        inserted_double_data[1]) - np.min(inserted_double_data[1])


def test_returns_max_minus_1_select(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select max(" + tslib._double_col_name(table) + ") - 1 from " + \
        table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    sanity_check(table.get_name(), len(inserted_double_data[0]), res, 1, 2)

    res_table = res.tables[table.get_name()]
    assert res_table[0].name == "$timestamp"
    assert res_table[0].data[0] >= start_time
    assert res_table[1].name == "(max(" + \
        tslib._double_col_name(table) + ")-1)"
    assert pytest.approx(res_table[1].data[0], 0.1) == np.max(
        inserted_double_data[1]) - 1


def test_returns_max_and_scalar_1_select(qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select max(" + tslib._double_col_name(table) + "), 1 from " + \
        table.get_name() + \
        " in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    assert len(res.tables) == 2

    res_table = res.tables["$none"]
    assert res_table[0].name == "$timestamp"
    assert np.isnat(res_table[0].data[0])
    assert res_table[1].name == "max(" + tslib._double_col_name(table) + ")"
    assert res_table[2].name == "1"
    assert res_table[2].data[0] == 1

    res_table = res.tables[table.get_name()]
    assert res_table[0].name == "$timestamp"
    assert res_table[0].data[0] >= start_time
    assert res_table[1].name == "max(" + tslib._double_col_name(table) + ")"
    assert pytest.approx(res_table[1].data[0], 0.1) == np.max(
        inserted_double_data[1])
    assert res_table[2].name == "1"


def test_returns_inserted_multi_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    inserted_blob_data = _insert_blob_points(table, start_time, 10)
    inserted_int64_data = _insert_int64_points(table, start_time, 10)
    inserted_timestamp_data = _insert_timestamp_points(table, start_time, 10)

    query = "select * from " + table.get_name() + \
        " in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query).run()

    # Column count is 5, because, uninit, int64, blob, timestamp, double
    sanity_check(table.get_name(), 4 * len(inserted_double_data[0]),
                 res, len(inserted_double_data[0]), 5)

    assert res.tables[table.get_name()][0].name == "$timestamp"
    assert res.tables[table.get_name(
    )][1].name == tslib._double_col_name(table)
    assert res.tables[table.get_name()][2].name == tslib._blob_col_name(table)
    assert res.tables[table.get_name()][3].name == tslib._int64_col_name(table)
    assert res.tables[table.get_name()][4].name == tslib._ts_col_name(table)

    np.testing.assert_array_equal(
        res.tables[table.get_name()][0].data, inserted_double_data[0])
    np.testing.assert_array_equal(
        res.tables[table.get_name()][1].data, inserted_double_data[1])
    np.testing.assert_array_equal(
        res.tables[table.get_name()][2].data, inserted_blob_data[1])
    np.testing.assert_array_equal(
        res.tables[table.get_name()][3].data, inserted_int64_data[1])
    np.testing.assert_array_equal(
        res.tables[table.get_name()][4].data, inserted_timestamp_data[1])



# # pylint: disable=C0103,C0111,C0302,W0212
# import os
# import sys
# import unittest
# import settings

# sys.path.append(os.path.join(os.path.split(__file__)[0], '..', 'examples'))

# class QuasardbExamples(unittest.TestCase):

#     def test_batch_ts_insert_example(self):
#         pass

#     def test_csv_insert_example(self):
#         HAS_PANDAS = True
#         try:
#             import pandas
#         except ImportError:
#             HAS_PANDAS = False
#         if HAS_PANDAS:
#             import csv_insert
#             csv_insert.main(settings.INSECURE_URI, settings.entry_gen.next(), os.path.join(
#                 os.path.split(__file__)[0], '..', 'examples', "fake_currency_minutes.csv"))
#         else:
#             print (
#                 "\nCannot run the example csv_insert as the module pandas is not found.\n")

#     def test_expiry_example(self):
#         import expiry
#         expiry.main(settings.INSECURE_URI, settings.entry_gen.next())

#     def test_query_tags_example(self):
#         import query_tags
#         query_tags.main(settings.INSECURE_URI)

#     def test_time_series_example(self):
#         import time_series
#         time_series.main(settings.INSECURE_URI, settings.entry_gen.next())

#     def test_temperature_example(self):
#         import temperature
#         temperature.main(settings.INSECURE_URI, settings.entry_gen.next(), 100)


# if __name__ == '__main__':
#     if settings.get_lock_status() is False:
#         settings.init()
#         test_directory = os.getcwd()
#         test_report_directory = os.path.join(os.path.split(
#             __file__)[0], '..', 'build', 'test', 'test-reports')
#         import xmlrunner
#         unittest.main(testRunner=xmlrunner.XMLTestRunner(  # pylint: disable=E1102
#             output=test_report_directory), exit=False)
#         settings.terminate()



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import quasardb


def _make_expiry_time(td):
    # expires in one minute
    now = datetime.datetime.now()
    # get rid of the microsecond for the testcases
    return now + td - datetime.timedelta(microseconds=now.microsecond)


def test_expires_at(blob_entry, random_blob):
    """
    Test for expiry.
    We want to make sure, in particular, that the conversion from Python datetime is right.
    """

    # entry does not exist yet
    with pytest.raises(quasardb.Error):
        blob_entry.get_expiry_time()

    blob_entry.put(random_blob)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert exp.year == 1970

    future_exp = _make_expiry_time(datetime.timedelta(minutes=1))
    blob_entry.expires_at(future_exp)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert exp == future_exp


def test_expires_from_now(blob_entry, random_blob):
    # entry does not exist yet
    with pytest.raises(quasardb.Error):
        blob_entry.get_expiry_time()

    blob_entry.put(random_blob)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert exp.year == 1970

    # expires in one minute from now
    blob_entry.expires_from_now(datetime.timedelta(minutes=1))

    # We use a wide 10s interval for the check, because we have no idea at which speed
    # these testcases may run in debug. This will be enough however to check that
    # the interval has properly been converted and the time zone is
    # correct.
    future_exp_lower_bound = datetime.datetime.now() + datetime.timedelta(seconds=50)
    future_exp_higher_bound = future_exp_lower_bound + \
        datetime.timedelta(seconds=80)

    exp = blob_entry.get_expiry_time()
    assert isinstance(exp, datetime.datetime)
    assert future_exp_lower_bound < exp
    assert future_exp_higher_bound > exp


def test_methods(blob_entry, random_blob):

    future_exp = _make_expiry_time(datetime.timedelta(minutes=1))

    blob_entry.put(random_blob, future_exp)

    exp = blob_entry.get_expiry_time()

    assert isinstance(exp, datetime.datetime)
    assert exp == future_exp

    future_exp = _make_expiry_time(datetime.timedelta(minutes=2))

    blob_entry.update(random_blob, future_exp)

    exp = blob_entry.get_expiry_time()

    assert isinstance(exp, datetime.datetime)
    assert exp == future_exp

    future_exp = _make_expiry_time(datetime.timedelta(minutes=3))

    blob_entry.get_and_update(random_blob, future_exp)

    exp = blob_entry.get_expiry_time()

    assert isinstance(exp, datetime.datetime)
    assert exp == future_exp

    future_exp = _make_expiry_time(datetime.timedelta(minutes=4))

    blob_entry.compare_and_swap(random_blob, random_blob, future_exp)

    exp = blob_entry.get_expiry_time()

    assert isinstance(exp, datetime.datetime)
    assert exp == future_exp



# Copyright (c) 2009-2019, quasardb SAS
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

import sys
import traceback
import os
import numpy as np

import quasardb  # pylint: disable=C0413,E0401

def main(uri_src, uri_dst, ts_name):

    print("Connecting to: ", uri_src)
    c_src = quasardb.Cluster(uri=uri_src)
    ts_src = c_src.ts(ts_name)

    print("Connecting to: ", uri_dst)
    c_dst = quasardb.Cluster(uri=uri_dst)
    ts_dst = c_dst.ts(ts_name)

    try:
        ts_dst.remove()
    except:
        pass

    # create the time series on the destination cluster.
    ts_dst.create([quasardb.ColumnInfo(quasardb.ColumnType.Int64, "bid"),
                   quasardb.ColumnInfo(quasardb.ColumnType.Int64, "last")])

    # Note that we are loading all data points of this entire column in
    # memory here. A more scalable solution would split the ranges into
    # smaller ranges, or use the streaming bulk reader and bulk inserter.
    everything = [(np.datetime64('1970-01-01', 'ns'), np.datetime64('2035-01-01', 'ns'))]

    data = ts_src.int64_get_ranges("bid", everything)
    ts_dst.int64_insert("bid", data[0], data[1])

    data = ts_src.int64_get_ranges("last", everything)
    ts_dst.int64_insert("last", data[0], data[1])

if __name__ == "__main__":

    try:
        if len(sys.argv) != 4:
            print("usage: ", sys.argv[0], " cluster_from cluster_to tsname")
            sys.exit(1)

        main(sys.argv[1], sys.argv[2], sys.argv[3])

    except Exception as ex:  # pylint: disable=W0703
        print("An error ocurred:", str(ex))
        traceback.print_exc()



# pylint: disable=C0103,C0111,C0302,W0212
import datetime


def test_get_timeout(qdbd_connection):
    qdbd_connection.options().set_timeout(datetime.timedelta(
        seconds=1))
    duration = qdbd_connection.options().get_timeout()

    assert duration == datetime.timedelta(seconds=1)



# pylint: disable=C0103,C0111,C0302,W0212
import numpy as np
import test_ts_batch as batchlib
import test_pandas as pandaslib
import quasardb.pandas as qdbpd

row_count = 10000


def test_bench_double_series(qdbd_connection, table, many_intervals, benchmark):
    batch_inserter = qdbd_connection.ts_batch(
        batchlib._make_ts_batch_info(table))

    doubles, blobs, integers, timestamps = batchlib._test_with_table(
        batch_inserter,
        table,
        many_intervals,
        batchlib._row_insertion_method,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_double")


def test_bench_blob_series(qdbd_connection, table, many_intervals, benchmark):
    batch_inserter = qdbd_connection.ts_batch(
        batchlib._make_ts_batch_info(table))

    doubles, blobs, integers, timestamps = batchlib._test_with_table(
        batch_inserter,
        table,
        many_intervals,
        batchlib._row_insertion_method,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_blob")


def test_bench_int64_series(qdbd_connection, table, many_intervals, benchmark):
    batch_inserter = qdbd_connection.ts_batch(
        batchlib._make_ts_batch_info(table))

    doubles, blobs, integers, timestamps = batchlib._test_with_table(
        batch_inserter,
        table,
        many_intervals,
        batchlib._row_insertion_method,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_int64")


def test_bench_timestamp_series(qdbd_connection, table, many_intervals, benchmark):
    batch_inserter = qdbd_connection.ts_batch(
        batchlib._make_ts_batch_info(table))

    doubles, blobs, integers, timestamps = batchlib._test_with_table(
        batch_inserter,
        table,
        many_intervals,
        batchlib._row_insertion_method,
        batchlib._regular_push)

    benchmark(qdbpd.read_series, table, "the_ts")


def test_benchmark_dataframe_read(qdbd_connection, table, benchmark):
    df = pandaslib.gen_df(np.datetime64('2017-01-01'), row_count)
    qdbpd.write_dataframe(df, qdbd_connection, table)
    benchmark(qdbpd.read_dataframe, table)


def test_benchmark_dataframe_write(qdbd_connection, table, benchmark):
    # Ensures that we can do a full-circle write and read of a dataframe
    df = pandaslib.gen_df(np.datetime64('2017-01-01'), row_count)
    benchmark(qdbpd.write_dataframe, df, qdbd_connection, table)



# pylint: disable=C0103,C0111,C0302,W0212,W0702

import pytest
import quasardb


def test_compression_none(qdbd_connection):
    qdbd_connection.options().set_compression(
        quasardb.Options.Compression.Disabled)


def test_compression_fast(qdbd_connection):
    qdbd_connection.options().set_compression(quasardb.Options.Compression.Fast)


def test_compression_invalid(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.options().set_compression(123)



# pylint: disable=C0103,C0111,C0302,R0903

# Copyright (c) 2009-2019, quasardb SAS
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import quasardb


class PandasRequired(ImportError):
    pass


try:
    import numpy as np
    import pandas as pd
    from pandas.core.api import DataFrame, Series
    from pandas.core.base import PandasObject
except ImportError:
    raise PandasRequired(
        "The pandas library is required to handle pandas data formats")


# Constant mapping of numpy dtype to QuasarDB column type
# TODO(leon): support this natively in qdb C api ? we have everything we need
#             to understand dtypes.
_dtype_map = {
    np.dtype('int64'): quasardb.ColumnType.Int64,
    np.dtype('int32'): quasardb.ColumnType.Int64,
    np.dtype('float64'): quasardb.ColumnType.Double,
    np.dtype('object'): quasardb.ColumnType.Blob,
    np.dtype('M8[ns]'): quasardb.ColumnType.Timestamp
}


def read_series(table, col_name, ranges=None):
    """
    Read a Pandas Timeseries from a single column.

    Parameters:
    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.ts('my_table')

    col_name : str
      Name of the column to read.

    ranges : list
      A list of ranges to read, represented as tuples of Numpy datetime64[ns] objects.
    """
    read_with = {
        quasardb.ColumnType.Double: table.double_get_ranges,
        quasardb.ColumnType.Blob: table.blob_get_ranges,
        quasardb.ColumnType.Int64: table.int64_get_ranges,
        quasardb.ColumnType.Timestamp: table.timestamp_get_ranges,
    }

    kwargs = {
        'column': col_name
    }

    if ranges:
        kwargs['ranges'] = ranges

    # Dispatch based on column type
    t = table.column_type_by_id(col_name)
    res = (read_with[t])(**kwargs)

    return Series(res[1], index=res[0])


def write_series(series, table, col_name):
    """
    Writes a Pandas Timeseries to a single column.

    Parameters:
    series : pandas.Series
      Pandas Series, with a numpy.datetime64[ns] as index. Underlying data will be attempted
      to be transformed to appropriate QuasarDB type.

    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.ts('my_table')

    col_name : str
      Column name to store data in.
    """
    write_with = {
        quasardb.ColumnType.Double: table.double_insert,
        quasardb.ColumnType.Blob: table.blob_insert,
        quasardb.ColumnType.Int64: table.int64_insert,
        quasardb.ColumnType.Timestamp: table.timestamp_insert
    }

    t = table.column_type_by_id(col_name)
    (write_with[t])(col_name, series.index.to_numpy(), series.to_numpy())


def read_dataframe(table, row_index=False, columns=None, ranges=None):
    """
    Read a Pandas Dataframe from a QuasarDB Timeseries table.

    Parameters:
    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.ts('my_table')

    columns : optional list
      List of columns to read in dataframe. The timestamp column '$timestamp' is
      always read.

      Defaults to all columns.

    row_index: boolean
      Whether or not to index by rows rather than timestamps. Set to true if your
      dataset may contains null values and multiple rows may use the same timestamps.
      Note: using row_index is significantly slower.
      Defaults to false.

    ranges: optional list
      A list of time ranges to read, represented as tuples of Numpy datetime64[ns] objects.
      Defaults to the entire table.

    """

    if columns is None:
        columns = list(c.name for c in table.list_columns())

    if not row_index:
        xs = dict((c, (read_series(table, c, ranges))) for c in columns)
        return DataFrame(data=xs)

    kwargs = {
        'columns': columns
    }
    if ranges:
        kwargs['ranges'] = ranges

    reader = table.reader(**kwargs)
    xs = []
    for row in reader:
        xs.append(row.copy())

    columns.insert(0, '$timestamp')
    return DataFrame(data=xs, columns=columns)


def write_dataframe(df, cluster, table, create=False, _async=False, chunk_size=50000):
    """
    Store a dataframe into a table.

    Parameters:
    df: pandas.DataFrame
      The pandas dataframe to store.

    cluster: quasardb.Cluster
      Active connection to the QuasarDB cluster

    table: quasardb.Timeseries or str
      Either a string or a reference to a QuasarDB Timeseries table object.
      For example, 'my_table' or cluster.ts('my_table') are both valid values.

    create: optional bool
      Whether to create the table. Defaults to false.
    """

    # Acquire reference to table if string is provided
    if isinstance(table, str):
        table = cluster.ts(table)

    if create:
        _create_table_from_df(df, table)

    # Create batch column info from dataframe
    col_info = list(quasardb.BatchColumnInfo(
        table.get_name(), c, chunk_size) for c in df.columns)
    batch = cluster.ts_batch(col_info)

    write_with = {
        quasardb.ColumnType.Double: batch.set_double,
        quasardb.ColumnType.Blob: batch.set_blob,
        quasardb.ColumnType.Int64: batch.set_int64,
        quasardb.ColumnType.Timestamp: lambda i, x: batch.set_timestamp(
            i, np.datetime64(x, 'ns'))
    }

    # Split the dataframe in chunks that equal our batch size
    dfs = [df[i:i+chunk_size] for i in range(0, df.shape[0], chunk_size)]

    for current_df in dfs:
        # TODO(leon): use pinned columns so we can write entire numpy arrays
        for row in current_df.itertuples(index=True):
            batch.start_row(np.datetime64(row[0], 'ns'))

            for i in range(len(current_df.columns)):
                v = row[i + 1]

                if not pd.isnull(v):
                    ct = _dtype_to_column_type(
                        current_df[current_df.columns[i]].dtype)
                    fn = write_with[ct]
                    fn(i, v)

        if _async is True:
            batch.push_async()
        else:
            batch.push()


def _create_table_from_df(df, table):
    cols = list(quasardb.ColumnInfo(
        _dtype_to_column_type(df[c].dtype), c) for c in df.columns)
    table.create(cols)
    return table


def _dtype_to_column_type(dt):
    res = _dtype_map.get(dt, None)
    if res is None:
        raise ValueError("Incompatible data type: ", dt)

    return res



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import quasardb


def test_set_timeout_1_day(qdbd_connection):
    qdbd_connection.options().set_timeout(datetime.timedelta(days=1))


def test_set_timeout_1_second(qdbd_connection):
    qdbd_connection.options().set_timeout(datetime.timedelta(seconds=1))


def test_set_timeout_throws__when_timeout_is_1_microsecond(qdbd_connection):
    with pytest.raises(quasardb.Error):
        qdbd_connection.options().set_timeout(datetime.timedelta(microseconds=1))
